import ast


def stringfy(node, inner=True):
    if isinstance(node, ast.Name):
        return str(node.id)
    if isinstance(node, ast.Attribute):
        return stringfy(node.value) + '.' + str(node.attr)
    if isinstance(node, ast.Call):
        return stringfy(node.func) + '()'
    if isinstance(node, ast.Subscript):
        return stringfy(node.value)+'['+stringfy(node.slice)+']'
    if isinstance(node, ast.Slice):
        if not inner:
            return '@'
        base = stringfy(node.lower)+':'+stringfy(node.upper)
        if node.step is None:
            return base
        return base + ':' + stringfy(node.step)
    if isinstance(node, ast.Index):
        return stringfy(node.value)
    if inner:
        if isinstance(node, ast.Constant):
            return str(node.value)
        if isinstance(node, ast.Num):
            return str(node.n)
        if isinstance(node, ast.Str):
            return str(node.s)
        if node is None:
            return ''
    return '@'


def traverse(node, ss: set):
    #print(ast.dump(node), stringfy(node, inner=False))
    snipet = stringfy(node, inner=False)
    if '@' not in snipet:
        ss.add(snipet)
    for sub_node in ast.iter_child_nodes(node):
        traverse(sub_node, ss)
    return ss


def extract_vars(code):
    try:
        node = ast.parse(code)
        ss = traverse(node, set())
        return [s for s in ss if (not s.endswith('()')) and (s+'()' not in ss)]
    except SyntaxError:
        return []



import ast
from IPython import get_ipython


def stringfy(node, inner=True):
    if isinstance(node, ast.Name):
        return str(node.id)
    if isinstance(node, ast.Attribute):
        return stringfy(node.value) + '.' + str(node.attr)
    if isinstance(node, ast.Call):
        return stringfy(node.func) + '()'
    if isinstance(node, ast.Subscript):
        return stringfy(node.value)+'['+stringfy(node.slice)+']'
    if isinstance(node, ast.Slice):
        if not inner:
            return '@'
        base = stringfy(node.lower)+':'+stringfy(node.upper)
        if node.step is None:
            return base
        return base + ':' + stringfy(node.step)
    if isinstance(node, ast.Index):
        return stringfy(node.value)
    if inner:
        if isinstance(node, ast.Constant):
            return str(node.value)
        if isinstance(node, ast.Num):
            return str(node.n)
        if isinstance(node, ast.Str):
            return str(node.s)
        if node is None:
            return ''
    return '@'


def traverse(node, ss: set):
    #print(ast.dump(node), stringfy(node, inner=False))
    snipet = stringfy(node, inner=False)
    if '@' not in snipet:
        ss.add(snipet)
    for sub_node in ast.iter_child_nodes(node):
        traverse(sub_node, ss)
    return ss


def extract_vars(code):
    try:
        node = ast.parse(code)
        ss = traverse(node, set())
        return [s for s in ss if (not s.endswith('()')) and (s+'()' not in ss)]
    except SyntaxError:
        return []


def eval_ipython(code=None, locals=None):
    ipy = get_ipython()
    globals = ipy.user_global_ns
    code = code or ipy.user_global_ns['In'][-1]
    locals = locals or ipy.user_ns
    vars = extract_vars(code)
    #print('@', code, vars)
    ss = []
    for snipet in vars:
        try:
            # print(snipet)
            v = eval(snipet, globals, locals)
            ss.append(dump_value(snipet, v))
        except BaseException as e:
            pass
            #ss.append((snipet, e))
    return ss

##
# dump


_PYTYPE = {
    'NoneType': 'None',
    'bool': 'ブール値',
    'int': '整数',
    'float': '浮動小数点数',
    'complex': '複素数',
    'str': '文字列',
    'bytes': 'バイト列',
    'list': 'リスト',
    'tuple': 'タプル',
    'dict': '辞書',
    'set': 'セット(集合)',
    'ndarray': '配列',
    'DataFrame': 'データフレーム',
    'module': 'モジュール',
    'builtin_function_or_method': 'ビルトイン関数',
    'function': '関数',
    'method': 'メソッド',
    'type': 'クラス(型)',
}


def _typename(value):
    typename = type(value).__name__
    if typename in _PYTYPE:
        return _PYTYPE[typename] + f'({typename}型)'
    return f'{typename}型'


def dump_value(key, value, html=False):
    ss = []
    ss.append(key)
    ss.append(_typename(value))
    if hasattr(value, 'shape'):
        ss.append(f'{key}.shape={value.shape}')
    elif hasattr(value, '__len__'):
        ss.append(f'len({key})={len(value)}')
    # if hasattr(value, '_repr_html_'):
    #     body = value._repr_html_()
    ss.append(repr(value))
    return ' '.join(ss)


def analyze_code(slots):
    code = slots.get('code', None)
    locals = slots.get('vars', None)
    ss = eval_ipython(code, locals)
    ss.insert(0, '変数の値を全部、出してみるよ（変な値はないか探してごらん)')
    #print('@', ss)
    if len(ss) > 1:
        slots['fault_vars'] = ss


def run_diagnosis(slots):
    if 'code' not in slots:
        try:
            slots['code'] = get_ipython().user_global_ns['In'][-1]
        except:
            pass
    if 'code' in slots and 'traceback' in slots:
        code = slots['code']
        lines = []
        for data in slots['traceback']:
            lineno = data['lineno']
            line = data['line'].strip()
            if line in code:
                line = f'[{lineno}行目] {line} に変なところない？'
                lines.append(line)
        if len(lines) > 0:
            slots['fault_lines'] = lines
    analyze_code(slots)
    # if 'problem_id' in slots:
    #     text = slots['problem_id']
    #     if text in HINT:
    #         text = HINT[text]
    #         slots['hint'] = text



import requests
from requests.exceptions import Timeout
from bs4 import BeautifulSoup

SAMPLE = {}


def download_atcoder_sample(url):
    if '?' in url:
        url, _, _ = url.rpartition('?')
    _, _, problem_id = url.rpartition('/')
    if problem_id in SAMPLE:
        return SAMPLE[problem_id]
    # try:
    response = requests.get(url, timeout=(3.0, 7.5))
    if response.status_code != 200:
        return {'error': f'読み込めません {response.status_code} {url}'}
    response_text = response.text
    html = BeautifulSoup(response_text, "lxml")
    d = {}
    for a in html.find_all("section"):
        # print(a)
        if a.h3 and a.pre:
            key = a.h3.text.replace('\r\n', '\n')
            value = a.pre.text.replace('\r\n', '\n')
            d[key] = value
    data = {'problem_id': problem_id, 'url': url}
    testcases = []
    if '入力例 1' in d:
        testcases.append(dict(input=d['入力例 1'], output=d['出力例 1']))
    if '入力例 2' in d:
        testcases.append(dict(input=d['入力例 2'], output=d['出力例 2']))
    if '入力例 3' in d:
        testcases.append(dict(input=d['入力例 3'], output=d['出力例 3']))
    data['testcases'] = testcases
    SAMPLE[problem_id] = data
    return data


# https://zenn.dev/hellorusk/articles/copy-from-blog-20210211
DROPBOX = 'https://www.dropbox.com/sh/nx3tnilzqz7df8a/AABP2YNyT09nYDTLV7d9bPwka/ABC102/A/in/sample_01.txt'
HOST = 'https://www.dropbox.com/sh/nx3tnilzqz7df8a/AACe0splFqRfbrMiRu41zmLna'


def download_atcoder_problem(directive):
    ss = directive.split()
    try:
        data = download_atcoder_sample(ss[0])
        return data
    except Timeout:
        return dict(error='ネットワーク接続できません')
    # problem_id = data['problem_id'].replace('_', '/').upper()
    # print(problem_id, ss)
    # for filename in ss[1:]:
    #     if not filename.endswith('.txt'):
    #         continue
    #     key = f'{problem_id}/{filename}'
    #     if key in SAMPLE:
    #         p = SAMPLE[key]
    #         data['testcases'].append(p)
    #         continue
    #     url = f'{HOST}/{problem_id}/in/{filename}?dl=0'
    #     print(url)
    #     response = requests.get(url, timeout=(3.0, 7.5))
    #     print(response.status_code)
    #     if response.status_code != 200:
    #         print('ファイルがありません', url)
    #         continue
    #     history = response.history
    #     if history:
    #         print('history:', history)
    #         print('history_url:', history[0].url)
    #         print('history_status_code:', history[0].status_code)
    #     input_text = response.text
    #     print(input_text)
    #     print(len(input_text))
    #     url = f'{HOST}/{problem_id}/out/{filename}?dl=0'
    #     print(url)
    #     response = requests.get(url, timeout=(3.0, 7.5))
    #     print(response.status_code)
    #     if response.status_code != 200:
    #         print('ファイルがありません', url)
    #         continue
    #     output_text = response.text
    #     p = dict(title=filename, input=input_text, output=output_text)
    #     SAMPLE[key] = p
    #     data['testcases'].append(p)
    # return data


if __name__ == '__main__':
    download_atcoder_problem(
        'https://atcoder.jp/contests/abc204/tasks/abc204_a random00.txt')



import os
from ._extract_emsg import extract_emsg, replace_eparams

EMSG_DIC = {}


def _define_emsg(key, lines):
    if key is None or len(lines) == 0:
        return
    if key in EMSG_DIC:
        d = EMSG_DIC[key]
    else:
        d = {}
        EMSG_DIC[key] = d
    for line in lines:
        key, _, value = line.partition(':')
        if key in d:
            d[key] = d[key] + '\n' + value.strip()
        else:
            d[key] = value.strip()


def _abspath(file):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), file)


def load_emsg_dic(file='emsg2_ja.txt'):
    if not os.path.exists(file):
        file = _abspath(file)

    with open(file) as f:
        ekey = None
        lines = []
        for line in f.readlines():
            line = line.strip()
            if line.startswith('#'):
                continue
            if line == '':
                _define_emsg(ekey, lines)
                ekey = None
                lines = []
                continue
            if ekey is None:
                ekey = line
            else:
                lines.append(line)
        _define_emsg(ekey, lines)


def _find_dic(emsg):
    ekey, params = extract_emsg(emsg)
    if ekey in EMSG_DIC:
        return ekey, params, EMSG_DIC[ekey]
    ekey2, params2 = extract_emsg(emsg, maybe=True)
    if ekey in EMSG_DIC:
        return ekey2, params2, EMSG_DIC[ekey]
    return ekey, params, None


def translate_emsg(emsg, record=None, translate_en=None):
    ekey, eparams, values = _find_dic(emsg)
    if record is None:
        record = {}

    record['ekey'] = ekey
    record['eparams'] = eparams
    # print(record)
    if values is None:
        if not translate_en:
            return None
        translated = translate_en(ekey)
        if not translated:
            return None
        record['translated'] = replace_eparams(translated, eparams)
        record['translated_by'] = 'TexTra'
    else:
        for key, value in values.items():
            record[key] = replace_eparams(value, eparams)
    return record['translated']


load_emsg_dic('emsg2_ja.txt')



import builtins
from kogi.dialog import kogi_catch
from .timeout import exec_with_timeout
from .judge import render_result, render_footer


def get_target(code):
    for c in code.splitlines()[::-1]:
        if '_ =' in c:
            return c.replace('_ =', '').strip()
        if '_=' in c:
            return c.replace('_=', '').strip()
    return ''


def safe_exec(code, globals, locals):
    try:
        exec(code, globals, locals)
    except Exception as e:
        locals['_'] = repr(e)


def judge_drill(code: str, data: dict, local_vars: dict):
    exec_with_timeout(code, None, local_vars, 15)
    if 'sample' in data:
        ref_vars = {}
        exec(data['sample'], None, ref_vars)
    else:
        ref_vars = local_vars.copy()

    for i, testcase in enumerate(data['testcases']):
        testcode = testcase.get('testcode')
        title = get_target(testcode)
        safe_exec(testcode, None, ref_vars)
        safe_exec(testcode, None, local_vars)
        sample = ref_vars['_']
        result = local_vars['_']
        render_result(title, '', result, sample)


# judge

_lines = None
_outputs = None


def judge_set(lines: list):
    global _lines, _outputs
    if lines is None:
        _lines = None
        _outputs = None
    else:
        _lines = lines[:]
        _outputs = []


def judge_input(prompt=''):
    global _lines
    if _lines is not None:
        if len(_lines) > 0:
            return _lines.pop(0)
        return ''
    return builtins.input(prompt)


def judge_print(*a, **kw):
    if _outputs is not None:
        sep = kw.get('sep', ' ')
        end = kw.get('end', '\n')
        s = sep.join([str(s) for s in a]) + end
        _outputs.append(s)
    else:
        builtins.print(*a, **kw)


def judge_cpc(ipy, code, data, context):
    context['problem_id'] = data['problem_id']
    global_vars = {
        'print': judge_print,
        'input': judge_input,
    }
    ac = 0
    for i, testcase in enumerate(data['testcases']):
        title = testcase.get('title', f'Case {i+1}')
        inputData = testcase['input']
        outputData = testcase['output']
        lines = [s for s in inputData.split('\n') if len(s) > 0]
        judge_set(lines)
        exec_with_timeout(code, global_vars, None, 10)
        resultData = ''.join(_outputs)
        ac += 1 if outputData == resultData else 0
        render_result(title, inputData, resultData, outputData)
    render_footer(data)


def kogi_judge(ipy, code, data, judge_fn):
    try:
        context = {}
        judge_fn(ipy, code, data, context)
    except SyntaxError as e:
        kogi_catch(code=code, context=context, exception=e)
    except:
        kogi_catch(code=code, context=context)



import os
import string

try:
    import pegtree as pg
except ModuleNotFoundError:
    os.system('pip install pegtree')
    import pegtree as pg


_PEG = '''

Start = { 
    (Param /  { (!Param .)* } )*
}

Param = LQuote / Quote / DoubleQuote / BackQuote 
    / Data / FuncName / MaybeName / CellName/ VarName 
    / ClassName / PathName / UName
    / Float / Int / Hex

LQuote = LQuote1 / LQuote2 / LQuote3
LQuote1 = {'\\'(' (!')\\'' . )+ ')\\'' #Quote}
LQuote2 = {'\\'<' (!'>\\'' . )+ '>\\'' #Quote}
LQuote3 = {'\\'[' ('\\\\' '\\'' / !']\\'' . )+ ']\\'' #Quote}
Quote = { '\\'' (!'\\'' .)* '\\'' #Quote }
BackQuote = { '`' (!'`' .)* '`' #Quote }
DoubleQuote = { '"' (![" ] .)* '"' #Quote }

Data = Set / Tuple
Set = { '{' ( Data / !'}' . )* '}' #Set }
Tuple = { '[' ( Data / !']' . )* ']' #Tuple }

FuncName = { NAME &'()' #FuncName }
CellName = { '%' '%'? NAME  #CellName }
VarName = { NAME ('\\'' NAME)? }  // can't
NAME = [A-Za-z_] [A-Za-z_.0-9]*
ClassName = { '<' [A-Za-z] (!'>' .)* '>' #ClassName }
PathName = { '(/' (!')' .)+ ')' #Path }

MaybeName = 
    / { NAME &(' object' !NAME) #Maybe }
    / { NAME &(' instance' !NAME) #Maybe }
    / { NAME &(' expected' !NAME) #Maybe }
    / { TYPENAME #Maybe }

TYPENAME =
    / 'list' !NAME
    / 'tuple' !NAME
    / 'int' !NAME
    / 'float' !NAME
    / 'str' !NAME
    / 'deque' !NAME

Float = { '-'? [0-9]* '.' [0-9]+ #Number }
Int = { '-'? [0-9]+ ('.py')? ![A-Za-z] #Int }
Hex = { '0x' [0-9A-Fa-f]+ #Hex }

UName = { U (!END .)* #UName }
END = [ (),]

U = [ぁ-んァ-ヶ㐀-䶵一-龠々〇〻ー]

'''

_parser = pg.generate(pg.grammar(_PEG))
_IDX = string.ascii_uppercase


def extract_params_from_error(emsg, maybe=False):
    tree = _parser(emsg)
    ss = []
    params = []
    for t in tree:
        s = str(t)
        if t == '':
            ss.append(s)
            continue
        if t == 'Maybe' and not maybe:
            ss.append(s)
            continue
        idx = _IDX[len(params) % 26]
        ss.append(idx)
        params.append(s)
    return ''.join(ss), params


def _safe(s):
    ss = list(s)
    for i in range(len(ss)-1):
        if 'A' <= ss[i] <= 'Z':
            c = ss[i+1]
            if 'A' <= c <= 'z' or '0' <= c <= '9' or c == '_':
                pass
            else:
                ss[i] = f'_{ss[i]}_'
    return ''.join(ss)


UNQUOTE_FORMAT = '{}'


def _unquote(s):
    if s[0] == s[-1] and s[0] == "'" or s[0] == '`':
        s2 = s[1:-1]
        for c in s2:
            if ord(c) > 127 or not c.isalnum() and c != '_':
                return s
        return UNQUOTE_FORMAT.format(s2)
    return UNQUOTE_FORMAT.format(s)


def replace_params(t, params):
    t = _safe(t)
    for X, val in zip(string.ascii_uppercase, params):
        t = t.replace(f'_{X}_', _unquote(val))
    return t


def abspath(file):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), file)


class ErrorModel(object):
    def __init__(self, file=None):
        self.eDict = {}
        if file is not None:
            self.load(file)

    def load(self, file='emsg_ja.txt'):
        if not os.path.exists(file):
            file = abspath(file)
        with open(file) as f:
            ekey = None
            lines = []
            for line in f.readlines():
                line = line.strip()
                if line.startswith('#'):
                    continue
                if line == '':
                    self.define_emsg(ekey, lines)
                    ekey = None
                    lines = []
                    continue
                if ekey is None:
                    ekey = line
                else:
                    lines.append(line)
            self.define_emsg(ekey, lines)

    def define_emsg(self, key, lines):
        if key is None or len(lines) == 0:
            return
        if key in self.eDict:
            d = self.eDict[key]
        else:
            d = {}
            self.eDict[key] = d
        for line in lines:
            key, _, value = line.partition(':')
            if key in d:
                d[key] = d[key] + '\n' + value.strip()
            else:
                d[key] = value.strip()

    def find_new(self, series):
        ss = set()
        for emsg in series:  # df['emsg']
            emsg = emsg.strip()
            emsg, params = extract_params_from_error(emsg)
            # print(emsg)
            ss.add(emsg)
            #print('\t', params)
        for s in ss:
            if len(s) < 256:
                if s not in self.eDict:
                    print(s)
        #print(len(series), '=>', len(ss))

    def translate(self, emsg):
        ekey, params = extract_params_from_error(emsg)
        if ekey in self.eDict:
            d = self.eDict[ekey]
            if 'translated' in d:
                return replace_params(d['translated'], params)
        return emsg

    def get_slots(self, emsg):
        ekey, params = extract_params_from_error(emsg)
        slots = {}
        if ekey in self.eDict:
            d = self.eDict[ekey]
            for key, value in d.items():
                slots[key] = replace_params(value, params)
            return slots
        ekey2, params2 = extract_params_from_error(emsg, maybe=True)
        if ekey2 in self.eDict:
            d = self.eDict[ekey2]
            for key, value in d.items():
                slots[key] = replace_params(value, params2)
        else:
            slots['error_key'] = ekey
            slots['error_params'] = params
            #print('@', slots)
        return slots



# translate
import builtins
import warnings
import IPython
from IPython.display import display, HTML
import os

from .logger import logging_asjson, send_log, print_nop

TRANSLATE_CSS_HTML = '''
<style>
.parent {
  background-color: #edebeb;
  width: 100%;
  height: 150px;
}
textarea {
  width: 100%;
  box-sizing: border-box;  /* ※これがないと横にはみ出る */
  height:120px;
  font-size: large;
  outline: none;           /* ※ブラウザが標準で付加する線を消したいとき */
  resize: none;
}
.box11{
//    padding: 0.5em 1em;
//    margin: 2em 0;
    color: #5d627b;
    background: white;
    border-top: solid 5px #5d627b;
    box-shadow: 0 3px 5px rgba(0, 0, 0, 0.22);
}
.box18{
  //padding: 0.2em 0.5em;
  //margin: 2em 0;
  color: #565656;
  background: #ffeaea;
  //background-image: url(https://2.bp.blogspot.com/-u7NQvQSgyAY/Ur1HXta5W7I/AAAAAAAAcfE/omW7_szrzao/s800/dog_corgi.png);
  background-size: 150%;
  background-repeat: no-repeat;
  background-position: top right;
  background-color:rgba(255,255,255,0.8);
  background-blend-mode:lighten;
  //box-shadow: 0px 0px 0px 10px #ffeaea;
  border: dashed 2px #ffc3c3;
  //border-radius: 8px;
}
.box16{
    //padding: 0.5em 1em;
    //margin: 2em 0;
    background: -webkit-repeating-linear-gradient(-45deg, #f0f8ff, #f0f8ff 3px,#e9f4ff 3px, #e9f4ff 7px);
    background: repeating-linear-gradient(-45deg, #f0f8ff, #f0f8ff 3px,#e9f4ff 3px, #e9f4ff 7px);
}
.box24 {
    position: relative;
    padding: 0.5em 0.7em;
    margin: 2em 0;
    background: #6f4b3e;
    color: white;
    font-weight: bold;
}
.box24:after {
    position: absolute;
    content: '';
    top: 100%;
    left: 30px;
    border: 15px solid transparent;
    border-top: 15px solid #6f4b3e;
    width: 0;
    height: 0;
}
</style>

<div class="parent">
<div style="float: left; width: 48%; text-align: right;">
<label class="box24" for="input">INPUT</label>
<textarea id="input" class="box16"></textarea>
</div>
<div style="float: left; width: 48%; text-align: right;">
<label class="box24" for="outout">OUTPUT</label>
<textarea id="output" class="box18 python" readonly></textarea>
</div>
</div>
<div id="js-loader" class="loader"></div>
'''

TRANSLATE_SCRIPT = '''
<script>
    var timer = null;
    var logtimer = null;
    var inputPane = document.getElementById('input');
    inputPane.addEventListener('input', (e) => {
        var text = e.srcElement.value;
        if(timer !== null) {
            clearTimeout(timer);
        }
        if(logtimer !== null) {
            clearTimeout(logtimer);
        }
        timer = setTimeout(() => {
            timer = null;
            (async function() {
                const result = await google.colab.kernel.invokeFunction('notebook.Convert', [text], {});
                const data = result.data['application/json'];
                const textarea = document.getElementById('output');
                textarea.textContent = data.result;
            })();
        }, 600);  // 何も打たななかったら600ms秒後に送信
        logtimer = setTimeout(() => {
            // logtimer = null;
            google.colab.kernel.invokeFunction('notebook.Logger', [], {});
        }, 60*1000*5); // 5分に１回まとめて送信
    });
</script>
'''


def check_sentencepiece():
    try:
        import sentencepiece
    except:
        print('Installing sentencepiece')
        os.system('pip install -q sentencepiece')
    try:
        import transformers
    except:
        print('Installing transformers')
        os.system('pip install -q transformers')


def load_mt5(model_id, qint8=True, device='cpu', log_class=None, print=print):
    check_sentencepiece()
    import torch
    from transformers import MT5ForConditionalGeneration, MT5Tokenizer
    model = MT5ForConditionalGeneration.from_pretrained(model_id)
    tokenizer = MT5Tokenizer.from_pretrained(model_id, is_fast=True)

    if qint8:
        model = torch.quantization.quantize_dynamic(
            model, {torch.nn.Linear}, dtype=torch.qint8
        )

    if isinstance(device, str):
        device = torch.device(device)
    model.to(device)

    def gready_search(s: str, max_length=128, beam=5) -> str:
        input_ids = tokenizer.encode_plus(
            s,
            add_special_tokens=True,
            max_length=max_length,
            padding="do_not_pad",
            truncation=True,
            return_tensors='pt').input_ids.to(device)
        if beam == 1:  # greedy_search
            greedy_output = model.generate(input_ids, max_length=max_length)
            t = tokenizer.decode(greedy_output[0], skip_special_tokens=True)
            if log_class is not None:
                logging_asjson('nmt', right_now=True,
                               mode_id=model_id,
                               log_class=log_class,
                               input=s,
                               output=t,
                               )
            return t
        # beem_search
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            outputs = model.generate(
                input_ids,
                # max_length=max_length,
                return_dict_in_generate=True, output_scores=True,
                temperature=1.0,          # 生成にランダム性を入れる温度パラメータ
                diversity_penalty=1.0,    # 生成結果の多様性を生み出すためのペナルティ
                num_beams=beam,
                #            no_repeat_ngram_size=2,
                num_beam_groups=beam,
                num_return_sequences=beam,
                repetition_penalty=1.5,   # 同じ文の繰り返し（モード崩壊）へのペナルティ
                early_stopping=True
            )
            results = [tokenizer.decode(out, skip_special_tokens=True)
                       for out in outputs.sequences]
            scores = [float(x) for x in outputs.sequences_scores]
            return results, scores
    return gready_search


_kogi_nmt_fn = None


def kogi_nmt_wakeup(model_id='kkuramitsu/kogi-mt5-test'):
    global _kogi_nmt_fn
    _kogi_nmt_fn = load_mt5(model_id, log_class='wakeup')


def kogi_nmt_talk(s: str, beam=1):
    if _kogi_nmt_fn is not None:
        return _kogi_nmt_fn(s, beam=beam)
    return None


def _transform_nop(text: str):
    return text


def nmt(model_id, load_nmt=load_mt5, log_class=None, kogi_mode=False,
        beam=1, device='cpu', qint8=True,
        transform_before=_transform_nop, transform_after=_transform_nop,
        input='入力', output='予測', print=print):
    global _kogi_nmt_fn
    nmt_fn = load_nmt(model_id, qint8=qint8, device=device,
                      log_class=log_class, print=print)
    _kogi_nmt_fn = nmt_fn

    cached = {'': ''}

    def convert(text):
        try:
            ss = []
            for line in text.splitlines():
                if line not in cached:
                    line = transform_before(line)
                    translated = nmt_fn(line, beam=beam)
                    translated = transform_after(translated)
                    print(len(line), line, '=>', translated)
                    cached[line] = translated
                else:
                    translated = cached[line]
                ss.append(translated)
            text = '\n'.join(ss)
            return IPython.display.JSON({'result': text})
        except Exception as e:
            print(e)

    try:
        display(HTML(TRANSLATE_CSS_HTML.replace(
            'INPUT', input).replace('OUTPUT', output)))
        display(HTML(TRANSLATE_SCRIPT))
        from google.colab import output
        output.register_callback('notebook.Convert', convert)
        output.register_callback('notebook.Logger', send_log)
    except ModuleNotFoundError as e:
        builtins.print('申し訳ありません. 現在、Colab上のみ動作します。')
        print(e)


def kogi_nmt(model_id, load_nmt=load_mt5, class_name='unknown',
             beam=1, device='cpu', qint8=True, print=print_nop):
    nmt(model_id, load_nmt=load_nmt, log_class=class_name, kogi_mode=True,
        beam=beam, device=device, qint8=qint8, input='日本語', output='Python', print=print)



def _ZEN():
    _ZA = ''.join(chr(ord('Ａ')+c) for c in range(26))
    _Za = ''.join(chr(ord('ａ')+c) for c in range(26))
    _Z0 = ''.join(chr(ord('０')+c) for c in range(10))
    _Z = '\u3000「」！＂＃＄％＆＇（）＊＋，－．／：；＜＝＞＠［＼］＾＿｀｛｜｝'
    return _Z + _ZA + _Za + _Z0


def _HAN():
    _HA = ''.join(chr(ord('A')+c) for c in range(26))
    _Ha = ''.join(chr(ord('a')+c) for c in range(26))
    _H0 = ''.join(chr(ord('0')+c) for c in range(10))
    _H = '\u0020\'\'!"#$%&\'()*+, -./:<=>@[\\]^_`{|}'
    return _H + _HA + _Ha + _H0


_ZEN2HAN = str.maketrans(_ZEN(), _HAN())


def zen2han(s: str) -> str:
    return s.translate(_ZEN2HAN)


def remove_suffixes(s, removed_suffixes):
    for suffix in removed_suffixes:
        if s.endswith(suffix):
            return remove_suffixes(s[:-len(suffix)], removed_suffixes)
    return s


REMOVED_SUFFIXES = [
    '.', '。', '?', '？', '！', '!',
    '何', '何ですか', '何でしょうか',
    'です'
]


def normalize(text):
    text = zen2han(text)
    return remove_suffixes(text, REMOVED_SUFFIXES)


def startswith(text, prefixes):
    for prefix in prefixes:
        if text.startswith(prefix):
            return True
    return False


def remove_tai(s):
    if s.endswith('したい'):
        return s[:-3]+'する'
    if s.endswith('きたい'):
        return s[:-3]+'く'
    if s.endswith('ちたい'):
        return s[:-3]+'つ'
    if s.endswith('にたい'):
        return s[:-3]+'ぬ'
    if s.endswith('りたい'):
        return s[:-3]+'る'
    if s.endswith('みたい'):
        return s[:-3]+'む'
    if s.endswith('いたい'):
        return s[:-3]+'う'
    if s.endswith('ぎたい'):
        return s[:-3]+'ぐ'
    if s.endswith('びたい'):
        return s[:-3]+'ぶ'
    return s[:-2]+'る'



from kogi.liberr.emodel import ErrorModel
import sys
import linecache
from numbers import Number
from .extract_vars import extract_vars


def bold(s):
    return f'\033[01m{s}\033[0m'


def glay(s):
    return f'\033[07m{s}\033[0m'


def red(s):
    return f'\033[31m{s}\033[0m'


def green(s):
    return f'\033[32m{s}\033[0m'


def yellow(s):
    return f'\033[33m{s}\033[0m'


def blue(s):
    return f'\033[34m{s}\033[0m'


def cyan(s):
    return f'\033[36m{s}\033[0m'


dots = green('...')


def _repr_list(value):
    if len(value) > 1:
        return f'[{value[0]}, {dots}]'
    return '[]'


REPR_VALUE = {

}


def kogi_register_repr(value_or_typename, repr_func):
    if not isinstance(value_or_typename, str):
        value_or_typename = type(value_or_typename).__name__
    REPR_VALUE[kogi_register_repr] = repr_func


def repr_value(value):
    typename = type(value).__name__
    if typename in REPR_VALUE:
        return REPR_VALUE[typename](value)
    if isinstance(value, str):
        s = repr(value)
        if len(s) > 32:
            s = s[:32] + dots
        return red(s)
    if isinstance(value, Number) or value is None:
        return repr(value)
    return cyan(f'({typename})')


def repr_vars(vars, start=None, end=None):
    ss = []
    for key, value in list(vars.items())[start:end]:
        if key.startswith('_'):
            continue
        value = repr_value(value)
        if value is not None:
            ss.append(f'{bold(key)}={value}')
    return ' '.join(ss)


def getline(filename, lines, n):
    if filename == '<string>':
        if 0 <= n-1 < len(lines):
            return lines[n-1]
        return ''
    if filename == '<unknown>':
        if 0 <= n-1 < len(lines):
            return lines[n-1]
        return ''
    return linecache.getline(filename, n).rstrip()


def arrow(lineno, here=False):
    s = str(lineno)
    if here:
        arrow = '-' * max(5-len(s), 0) + '> '
    else:
        arrow = ' ' * max(5-len(s), 0) + '  '
    return red(arrow) + green(f'{s}')


def filter_expressions(vars, exprs=None):
    if exprs is not None:
        newvars = {}
        for key, value in vars.items():
            if key in exprs:
                newvars[key] = value
        return newvars
    return vars


def print_func(filename, funcname, local_vars, exprs=None, n_args=0):
    if filename.startswith('<ipython-input-'):
        t = funcname.split('-')
        if len(t) > 2:
            filename = f'[{t[2]}]'
    arguments = repr_vars(local_vars, 0, n_args)
    if len(arguments) > 2:
        arguments = f'({arguments})'
    locals = repr_vars(filter_expressions(local_vars, exprs), n_args)
    if '/ipykernel_' in filename:
        print(f'{bold(funcname)}{arguments}')
    elif filename.endswith('.py'):
        print(f'"{glay(filename)}" {bold(funcname)}{arguments}')
        return
    else:
        print(f'"{glay(filename)}" {bold(funcname)}{arguments}')
    if len(locals) > 2:
        print(locals)


def print_linecode(filename, lines, lineno):
    if lineno-2 > 0:
        print(arrow(lineno-2), getline(filename, lines, lineno-2))
    if lineno-1 > 0:
        print(arrow(lineno-1), getline(filename, lines, lineno-1))
    line = getline(filename, lines, lineno)
    print(arrow(lineno, here=True), getline(filename, lines, lineno))
    print(arrow(lineno+1), getline(filename, lines, lineno+1))
    print(arrow(lineno+2), getline(filename, lines, lineno+2))
    return dict(filename=filename, lineno=lineno, line=line)


def print_header(etype):
    print(red('-'*79))
    etype = str(etype.__name__).ljust(46) + ' Traceback(most recent call last)'
    print(bold(red(etype)))


def print_syntax_error(exception, slots, logging_json=None):
    lines = slots['code'].splitlines()
    filename = exception.filename
    slots['lineno'] = lineno = exception.lineno
    slots['line'] = text = exception.text
    slots['offset'] = offset = exception.offset
    print_func(filename, f'[lineno: {lineno} offset: {offset}]', {})
    if lineno-2 > 0:
        print(arrow(lineno-2), getline(filename, lines, lineno-2))
    if lineno-1 > 0:
        print(arrow(lineno-1), getline(filename, lines, lineno-1))
    print(arrow(lineno, here=True), getline(filename, lines, lineno))
    offset = max(0, offset-1)
    print(arrow(lineno), ' '*offset+bold(red('^^')))
    print(f"{bold(red(exception.__class__.__name__))}: {bold(exception.msg)}")
    if logging_json is not None:
        logging_json(
            type='syntax_error',
            code=slots['code'],
            emsg=slots['emsg'],
            lineno=lineno, line=text, offset=offset
        )
    translate_error(slots, logging_json=logging_json)
    return slots


def print_tb(etype, evalue, tb, slots, logging_json=None):
    print_header(etype)

    code = slots['code']
    lines = code.splitlines()
    if code != '':
        exprs = extract_vars(code)
        slots['exprs_in_code'] = exprs
        exprs = set(exprs)

    prev = None
    repeated = 0
    stacks = []
    while tb:
        filename = tb.tb_frame.f_code.co_filename
        if '-packages/' not in filename:
            funcname = tb.tb_frame.f_code.co_name
            n_args = tb.tb_frame.f_code.co_argcount
            lineno = tb.tb_lineno
            local_vars = tb.tb_frame.f_locals
            cur = (filename, funcname, lineno)
            if cur != prev:
                if repeated > 10:
                    print(f'... repeated {red(str(repeated))} times ...')
                print_func(filename, funcname, local_vars, exprs, n_args)
                stack = print_linecode(filename, lines, lineno)
                stacks.append(stack)
                repeated = 0
            else:
                if repeated < 10:
                    print_func(filename, funcname, local_vars, exprs, n_args)
                repeated += 1
            prev = cur
        tb = tb.tb_next
    slots['traceback'] = list(stacks[::-1])
    print(f"{bold(red(etype.__name__))}: {bold(evalue)}")
    if logging_json is not None:
        logging_json(
            type='runtime_error',
            code=slots['code'],
            emsg=slots['emsg'],
            traceback=list(stacks[::-1])
        )
    translate_error(slots, logging_json=logging_json)
    return slots


defaultErrorModel = ErrorModel('emsg_ja.txt')


def translate_error(slots, logging_json=None):
    slots.update(defaultErrorModel.get_slots(slots['emsg']))
    if logging_json is not None and 'error_key' in slots:
        logging_json(
            type='unknown_error',
            emsg=slots['emsg'],
            error_key=slots['error_key'],
            error_params=slots['error_params']
        )


def kogi_print_exc(code='', exc_info=None, exception=None, logging_json=None):
    if exc_info is None:
        etype, evalue, tb = sys.exc_info()
    else:
        etype, evalue, tb = exc_info
    if etype is None:
        return None
    slots = dict(
        etype=f'{etype.__name__}',
        emsg=(f'{etype.__name__}: {evalue}').strip(),
        code=code,
    )
    if isinstance(exception, SyntaxError):
        return print_syntax_error(exception, slots, logging_json=logging_json)
    if exception is None and issubclass(etype, SyntaxError):
        try:
            raise
        except SyntaxError as e:
            exception = e
        return print_syntax_error(exception, slots, logging_json=logging_json)
    return print_tb(etype, evalue, tb, slots, logging_json=logging_json)




from .settings import kogi_set, kogi_print
from .ui.slides import slide

set = kogi_set
print = kogi_print

try:
    from .exception_hook import enable_kogi_hook, kogi_register_hook
    enable_kogi_hook()
    from .problem import atcoder_detector, atcoder_judge
    kogi_register_hook('atcoder', atcoder_judge, atcoder_detector)
    from .ui import rmt
except ModuleNotFoundError as e:
    import traceback
    traceback.print_exc()
    kogi_print('Only Available on Google Colab')
    pass



import traceback
import uuid
import json
import signal
from datetime import datetime
import requests


def kogi_print(*args, **kw):
    print('\033[34m[🐶]', *args, **kw)
    print('\033[0m', end='')


def print_nop(*x, **kw):
    pass


# Slack Messaging

slack_conn = None
HOST = 'slack.com'


def _connect_slack():
    global slack
    if slack_conn is not None:
        return slack_conn
    slack_id = kogi_get('slack_id')
    if slack_id is None:
        return None
    try:
        from slackweb import Slack
    except ModuleNotFoundError:
        import os
        os.system('pip install slackweb')
        from slackweb import Slack
    url = f'https://hooks.{HOST}/services/{slack_id}'
    try:
        slack_conn = Slack(url)
        return slack_conn
    except Exception as e:
        kogi_print('Slack Error', e)
    return None


def send_message(text):
    try:
        conn = _connect_slack()
        if conn is None:
            return False
        conn.notify(text=text)
        return True
    except Exception as e:
        kogi_print('Slack Error:', e)
    return False


slack = None


def load_slack(slack_id='QNZDoUPuLo7C3lHyh9PWhZz3', update=True):
    global slack
    if slack_id is None:
        slack = None
        return None
    try:
        from slackweb import Slack
    except ModuleNotFoundError:
        import os
        os.system('pip install slackweb')
        from slackweb import Slack
    HOST = 'slack.com'
    ID = 'T02NYCBFP7B'
    ID2 = 'B02QPM8HNBH'
    if '/' not in slack_id:
        url = f'https://hooks.{HOST}/services/{ID}/{ID2}/{slack_id}'
    else:
        url = f'https://hooks.{HOST}/services/{slack_id}'
    try:
        local_slack = Slack(url)
        if update:
            slack = local_slack
    except Exception as e:
        print('Slack Error', e)
    return local_slack


def send_slack(logs, slack_id='QNZDoUPuLo7C3lHyh9PWhZz3'):
    try:
        local_slack = load_slack(slack_id=slack_id, update=False)
        jsondata = json.dumps(logs, ensure_ascii=False)
        local_slack.notify(text=jsondata)
    except Exception as e:
        print('Slack Error:', e)


SESSION = str(uuid.uuid1())
SEQ = 0
LOGS = []
UID = 'unknown'
POINT = 'ixe8peqfii'
HOST2 = 'amazonaws'
KEY = 'OjwoF3m0l20OFidHsRea3ptuQRfQL10ahbEtLa'
prev_epoch = datetime.now().timestamp()


def send_log(right_now=True, print=kogi_print):
    global prev_epoch, LOGS, POINT
    now = datetime.now().timestamp()
    delta = (now - prev_epoch)
    prev_epoch = now
    if len(LOGS) > 0 and (right_now or delta > 30 or len(LOGS) > 4):
        data = {
            "session": SESSION,
            "uid": UID,
            "logs": LOGS.copy(),
        }
        LOGS.clear()
        url = f'https://{POINT}.execute-api.ap-northeast-1.{HOST2}.com/dev'
        headers = {'x-api-key': f'A{KEY}s'}
        r = requests.post(url, headers=headers, json=data)
        #print('logging', r.status_code, data)
        if r.status_code != 200:
            print(r.status_code)
            print(r)
            #print(f'delta={delta} data={data}')


def log(**kw):
    global SEQ, LOGS, epoch
    now = datetime.now()
    date = now.isoformat(timespec='seconds')
    logdata = dict(seq=SEQ, date=date, **kw)
    LOGS.append(logdata)
    SEQ += 1
    send_log(right_now=False)
    return logdata


def logging_json(**kw):
    global SEQ, LOGS, epoch
    now = datetime.now()
    date = now.isoformat(timespec='seconds')
    logdata = dict(seq=SEQ, date=date, **kw)
    LOGS.append(logdata)
    SEQ += 1
    send_log(right_now=False)
    return logdata


def logging_asjson(log_type, right_now=False, **kwargs):
    global SEQ, LOGS, epoch
    now = datetime.now()
    date = now.isoformat(timespec='seconds')
    logdata = dict(log_type=log_type, seq=SEQ, date=date)
    logdata.update(kwargs)
    LOGS.append(logdata)
    SEQ += 1
    send_log(right_now=right_now)
    return logdata


LAZY_LOGGER = []


def add_lazy_logger(func):
    LAZY_LOGGER.append(func)


def sync_lazy_loggger():
    for logger in LAZY_LOGGER:
        try:
            logger()
        except:
            traceback.print_exc()


def _handler(signum, frame):
    sync_lazy_loggger()
    version = None
    try:
        import google.colab as colab
        version = f'colab {colab.__version__}'
    except ModuleNotFoundError:
        pass
    if version is None:
        version = 'unknown'
    logging_asjson('terminal', right_now=True, version=version)


signal.signal(signal.SIGTERM, _handler)



# kogi global settings

import os
import warnings
import json
import requests
from requests_oauthlib import OAuth1

from kogi.logger import logging_asjson, print_nop

GLOBALS = {
    # 'class': 'レギオ入門',
    # 'name': 'たぬき',
    'textra': 'cb25461ac40e7a2dc0b2bc05d381995a',
    'model_key': 'rhOcswxkXzMbhlkKQJfytbfxAPVsblhRHX',
}


def kogi_get(key, value=None):
    return GLOBALS.get(key, value)


def kogi_set(**kwargs):
    global GLOBALS
    GLOBALS.update(kwargs)
    if 'model_id' in kwargs:
        load_mt5(kwargs['model_id'])
    if 'textra_key' in kwargs:
        load_textra(kwargs['textra_key'])
    if 'slack_key' in kwargs:
        load_slack(kwargs['slack_key'])


# LOGGING

def kogi_print(*args, **kwargs):
    if GLOBALS.get('verbose', True):
        print('\033[35m[🐶]', *args, **kwargs)
        print('\033[0m', end='')


def kogi_log(log_type, right_now=True, **kwargs):
    global GLOBALS
    if 'class' not in kwargs:
        if 'class_name' not in GLOBALS:
            return
        kwargs['class'] = GLOBALS['class_name']
    if 'name' in GLOBALS:
        kwargs['user_name'] = GLOBALS['name']
    # kogi_print(kwargs)
    logging_asjson(log_type, right_now=right_now, **kwargs)


# Translate

TEXTRA_NAME = 'kkuramitsu'
TEXTRA_KEY = '6c0bbdfd6c5c53cb0b0699729ed56a5c062ebba7c'
TEXTRA_URL = 'https://mt-auto-minhon-mlt.ucri.jgn-x.jp/api/mt/generalNT'
TEXTRA_CACHE = {}
TexTraOAuth = None


def load_textra(secret):
    global TexTraOAuth
    TexTraOAuth = OAuth1(TEXTRA_KEY, secret)


def _isEnglish(text):
    for c in text:
        if c >= 'あ':
            return False
    return True


def translate(text, lang=None):
    global TexTraOAuth
    if TexTraOAuth is None:
        return None

    if lang is not None:
        URL = f'{TEXTRA_URL}_{lang}/'
    elif _isEnglish(text):
        URL = f'{TEXTRA_URL}_en_ja/'
    else:
        URL = f'{TEXTRA_URL}_ja_en/'

    isMulti = False
    if isinstance(text, list):
        text = '\n'.join(text)
        isMulti = True

    if text in TEXTRA_CACHE:
        return TEXTRA_CACHE[text]

    params = {
        'key': TEXTRA_KEY,
        'name': TEXTRA_NAME,
        'type': 'json',
        'text': text,
    }

    try:
        res = requests.post(URL, data=params, auth=TexTraOAuth)
        res.encoding = 'utf-8'
        data = json.loads(res.text)
        result = data['resultset']['result']['text']
        if isMulti:
            return result.split('\n')
        TEXTRA_CACHE[text] = result
        return result
    except Exception as e:
        #     kogi_print('翻訳エラー:', e)
        return None


def translate_en(text):
    return translate(text, lang='en_ja')


def translate_ja(text):
    return translate(text, lang='ja_en')

# NMT


mt5_model_id = None
mt5_model = None
mt5_tokenizer = None


def check_sentencepiece():
    try:
        import sentencepiece
    except:
        kogi_print('Installing sentencepiece')
        os.system('pip install -q sentencepiece')
    try:
        import transformers
    except:
        kogi_print('Installing transformers')
        os.system('pip install -q transformers')


def load_mt5(model_id, qint8=True, device='cpu'):
    global mt5_model_id, mt5_model, mt5_tokenizer, mt5_device
    if model_id == mt5_model_id:
        return

    check_sentencepiece()
    import torch
    from transformers import MT5ForConditionalGeneration, MT5Tokenizer

    model_id = model_id
    model = MT5ForConditionalGeneration.from_pretrained(model_id)
    tokenizer = MT5Tokenizer.from_pretrained(model_id, is_fast=True)

    if qint8:
        model = torch.quantization.quantize_dynamic(
            model, {torch.nn.Linear}, dtype=torch.qint8
        )

    if isinstance(device, str):
        device = torch.device(device)
    model.to(device)

    mt5_model_id = model_id
    mt5_model = model
    mt5_tokenizer = tokenizer
    mt5_device = device


def generate_gready(s: str, max_length=128, print=print) -> str:
    global mt5_model, mt5_tokenizer, mt5_device
    input_ids = mt5_tokenizer.encode_plus(
        s,
        add_special_tokens=True,
        max_length=max_length,
        padding="do_not_pad",
        truncation=True,
        return_tensors='pt').input_ids.to(mt5_device)

    greedy_output = mt5_model.generate(input_ids, max_length=max_length)
    t = mt5_tokenizer.decode(greedy_output[0], skip_special_tokens=True)
    #kogi_log(type='nmt', mode_id=mt5_model_id, input=s, output=t)
    return t


def generate_beam(s: str, beam: int, max_length=12, print=print) -> str:
    global mt5_model, mt5_tokenizer, mt5_device
    input_ids = mt5_tokenizer.encode_plus(
        s,
        add_special_tokens=True,
        max_length=max_length,
        padding="do_not_pad",
        truncation=True,
        return_tensors='pt').input_ids.to(mt5_device)

    # beem_search
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', UserWarning)
        outputs = mt5_model.generate(
            input_ids,
            # max_length=max_length,
            return_dict_in_generate=True, output_scores=True,
            temperature=1.0,          # 生成にランダム性を入れる温度パラメータ
            diversity_penalty=1.0,    # 生成結果の多様性を生み出すためのペナルティ
            num_beams=beam,
            #            no_repeat_ngram_size=2,
            num_beam_groups=beam,
            num_return_sequences=beam,
            repetition_penalty=1.5,   # 同じ文の繰り返し（モード崩壊）へのペナルティ
            early_stopping=True
        )
        results = [mt5_tokenizer.decode(out, skip_special_tokens=True)
                   for out in outputs.sequences]
        scores = [float(x) for x in outputs.sequences_scores]
        return results, scores


API_URL = "https://api-inference.huggingface.co/models/kkuramitsu/kogi-mt5-test"
API_CACHE = {}


def generate_api(text, model_key, print=print):
    global API_CACHE
    if len(text) > 120:
        return 'ぐるるるる\n（入力が長すぎます）'
    if text in API_CACHE:
        return API_CACHE[text]
    payload = {"inputs": text}
    headers = {"Authorization": f"Bearer {model_key}"}
    response = requests.post(API_URL, headers=headers, json=payload)
    output = response.json()
    # print(text, type(output), output)
    if isinstance(output, (list, tuple)):
        output = output[0]
    if 'generated_text' in output:
        result = output['generated_text']
        API_CACHE[text] = result
    return 'ねむねむ。まだ、起きられない！\n（しばらく待ってからもう一度試してください）'


def model_generate(text, beam=1, max_length=128, print=print) -> str:
    global GLOBALS, mt5_model
    if mt5_model is not None:
        if beam > 1:
            return generate_beam(text, beam, max_length, print)
        return generate_gready(text, max_length, print)
    if 'model_key' in GLOBALS:
        model_key = GLOBALS['model_key']
        model_key = f'hf_{model_key}'
        return generate_api(text, model_key, print)
    return None

# Slack


SLACK_COM = None


def load_slack(slack_id):
    global SLACK_COM
    try:
        from slackweb import Slack
    except ModuleNotFoundError:
        import os
        os.system('pip install slackweb')
        from slackweb import Slack
    url = f'https://hooks.slack.com/services/{slack_id}'
    try:
        SLACK_COM = Slack(url)
    except Exception as e:
        kogi_print('Slackに接続できませんでした.', e)


def send_slack(text):
    global SLACK_COM
    if SLACK_COM is None:
        return
    try:
        SLACK_COM.notify(text=text)
    except Exception as e:
        kogi_print('Slack Error:', e)



import sys
import traceback
from IPython import get_ipython

from .settings import (
    model_generate, translate_en, translate_ja,
    send_slack, kogi_get, kogi_log, kogi_print
)

from .diagnosis import run_diagnosis
from .dialog_desc import get_desc

try:
    from google.colab import output
except ModuleNotFoundError:
    output = None

from kogi.ui import kogi_display, display_dialog, Conversation
from kogi.liberr import kogi_print_exc, replace_eparams
from .logger import add_lazy_logger

import kogi.fake_nlp as nlp

MODULES = [
    ('math.', 'import math'),
    ('os.', 'import os'),
    ('sys.', 'import sys'),
    ('collections.', 'import collections'),
    ('time.', 'import time'),
    ('datetime.', 'import datetime'),
    ('random.', 'import random'),
    ('np.', 'import numpy as np'),
    ('plt.', 'import matplotlib.pyplot as plt'),
    ('pd.', 'import pandas as pd'),
]


def check_module(code):
    for module, todo in MODULES:
        if code.startswith(module):
            return f'<tt>{todo}</tt> してから、'
    return ''


def response_codegen(text: str):
    response = model_generate(text)
    if response is None:
        return 'kogi.set(...)を再実行しよう'
    code = response.replace('<nl>', '\n').replace('<tab>', '    ')
    return check_module(code) + f'<pre>{code}</pre>'


def response_hint(slots: dict):
    if 'ekey' in slots and 'eparams' in slots:
        ekey = slots['ekey']
        eparams = ' '.join(slots['eparams'])
        eline = slots.get('eline', '')
        text = f'{ekey}<tab>{eparams}<tab>{eline}'
        ans = model_generate(text)
        if ans:
            # kogi_print(ans, slots['eparams'])
            return replace_eparams(ans, slots['eparams'])
            # return f'{ans}<br>{translate_ja(ans)}'
    return None


def response_talk(text: str):
    response = model_generate(f'talk: {text}')
    if response is None:
        return 'ZZ.. zzz.. 眠む眠む..'
    return response


def response_desc(text: str):
    response = get_desc(text)
    if response is None:
        return response_talk(text)
    return response


class Chatbot(Conversation):

    def response(self, user_input):
        text = user_input
        if 'user_inputs' not in self.slots:
            self.slots['user_inputs'] = []
        self.slots['user_inputs'].append(text)
        # if nlp.startswith(text, ('質問')):
        #     return self.response_question(text)
        text = nlp.normalize(user_input)
        # if nlp.startswith(text, ('デバッグ', '助けて', 'たすけて', '困った', '分析', '調べて')):
        #     if 'fault_vars' in self.slots:
        #         return self.slots['fault_vars']
        #     else:
        #         return 'コギーも助けて...'
        if text.endswith('には'):
            text = text[:-2]
            return response_codegen(text)
        if text.endswith('たい'):
            text = nlp.remove_tai(text)
            return response_codegen(text)
        if text.endswith('って') or text.endswith('とは'):
            text = text[:-2]
            return response_desc(text)
        if nlp.startswith(text, ('原因', '理由', 'なぜ', 'なんで', 'どう')):
            response = response_hint(self.slots)
            if response is None:
                if 'reason' in self.slots:
                    return self.slots['reason']
                if 'solution' in self.slots:
                    return self.slots['solution']
                if 'maybe' in self.slots:
                    return 'ひょっとしたら、' + self.slots['maybe']
                if 'ekey' in self.slots:
                    return 'エラーメッセージを検索してみたら？'
                return 'エラーなくない？'
            return response
        if nlp.startswith(text, ('ヒント', '助けて', 'たすけて')):
            if 'hint' in self.slots:
                return self.slots['hint']
            else:
                return 'ヒントなし'
        return response_talk(text)


if output is None:

    def show_slots(slots, print=kogi_display):
        if 'reason' in slots:
            print(slots['reason'])
        if 'fault_lines' in slots:
            for reason in slots['fault_lines']:
                print(reason)
        if 'solution' in slots:
            print(slots['solution'])
        if 'maybe' in slots:
            print(slots['maybe'])
        if 'fault_vars' in slots:
            for reason in slots['fault_vars']:
                print(reason)
        if 'hint' in slots:
            print(slots['hint'])

    def _start_chat(chatbot, start_message):
        try:
            chatbot.get('bot_name', 'コギー')
            kogi_display(start_message)
            show_slots(chatbot.slots)
        except:
            kogi_print('バグりました。ご迷惑をおかけします')
            traceback.print_exc()

    def _start_chat(chatbot, start_message):
        bot, _ = display_dialog(chatbot)
        bot(start_message)

else:
    def _start_chat(chatbot, start_message):
        bot, _ = display_dialog(chatbot)
        bot(start_message)


global_slots = {
    'bot_name': 'コギー',
    'your_name': 'あなた',
}


def set_global_slots(**kwargs):
    for key, value in kwargs.items():
        global_slots[key] = value


PREV_CHAT = None
CHAT_CNT = 0


def record_dialog():
    global PREV_CHAT, CHAT_CNT
    if PREV_CHAT is None:
        return
    chat = PREV_CHAT
    PREV_CHAT = None
    CHAT_CNT += 1
    user = kogi_get('name', 'ユーザ')
    data = {'type': 'kogi_chat'}
    data.update(chat.slots)
    data['chat'] = chat.records
    kogi_log('kogi_chat', right_now=True, **data)
    # Slack レポート
    lines = [f'*{user}({CHAT_CNT})より*']
    if 'code' in data:
        lines.extend(['```', data['code'], '```'])
    if 'emsg' in data:
        lines.extend([data['emsg'], ''])
    if 'start' in data:
        lines.extend([data['start']])
    #print(chat.records, data)
    if len(chat.records) > 0:
        for user_text, bot_text in chat.records:
            lines.extend([f'> {user_text}', bot_text])
    send_slack('\n'.join(lines))


add_lazy_logger(record_dialog)


def start_dialog(slots: dict):
    global PREV_CHAT
    record_dialog()
    dialog_slots = global_slots.copy()
    dialog_slots.update(slots)
    dialog_slots['your_name'] = kogi_get('name', 'あなた')
    if 'start' not in dialog_slots:
        dialog_slots['start'] = dialog_slots.get('translated', 'おはよう')
    PREV_CHAT = Chatbot(slots=dialog_slots)
    _start_chat(PREV_CHAT, dialog_slots['start'])
    return PREV_CHAT


def kogi_catch(exc_info=None, code: str = None, context: dict = None, exception=None, enable_dialog=True):
    if exc_info is None:
        exc_info = sys.exc_info()
    slots = kogi_print_exc(code=code,
                           exc_info=exc_info, caught_ex=exception,
                           translate_en=translate_en)
    if context is not None:
        slots.update(context)
    run_diagnosis(slots)
    if enable_dialog:
        start_dialog(slots)


if __name__ == '__main__':
    start_dialog({})



#ex1-1
import pandas as pd
df = pd.read_csv('pollen.csv')

# test
_ = df.head()

#ex1-2
import pandas as pd
df = pd.read_csv('pollen.csv')

# test
_ = df.sort_values(by='平均気温', ascending=False)


#ex1-3
import pandas as pd
df = pd.read_csv('pollen.csv')
plt.hist(df['杉花粉飛散量'], width=0.8, color='pink')

#test 
_ = tosvg(plt)




# import linecache
# import sys
# import json
# import traceback

# from kogi.liberr.emodel import ErrorModel, extract_params_from_error
# from kogi.liberr.print_tb import kogi_print_exc, kogi_register_repr

from ._extract_emsg import extract_emsg, replace_eparams
from ._translate import translate_emsg, load_emsg_dic
from ._traceback import kogi_print_exc, kogi_register_repr


# class KogiError(Exception):
#     def __init__(self, **kw):
#         Exception.__init__(self, json.dumps(kw))


# _defaultErrorModel = ErrorModel('emsg_ja.txt')
# def catch_exception(exc_info=None, code=None, include_translated=True, include_locals=False, local_vars={}, logging_json=None):
#     if exc_info is None:
#         etype, evalue, tb = sys.exc_info()
#     else:
#         etype, evalue, tb = exc_info
#     emsg = (f"{etype.__name__}: {evalue}").strip()

#     stacks = []
#     stack_vars = []
#     eline = None
#     elineno = None
#     while tb:
#         filename = tb.tb_frame.f_code.co_filename
#         if '<string>' in filename and code is not None:
#             name = tb.tb_frame.f_code.co_name
#             elineno = tb.tb_lineno
#             lines = code.splitlines()
#             eline = lines[elineno-1].strip()
#             stacks.append(dict(filename=filename,
#                                etype=f'{etype.__name__}', emsg=emsg,
#                                name=name, lineno=elineno, line=eline))
#             stack_vars.append(local_vars)
#         elif 'lib' not in filename:
#             name = tb.tb_frame.f_code.co_name
#             lineno = tb.tb_lineno
#             line = linecache.getline(filename, lineno, tb.tb_frame.f_globals)
#             local_vars = tb.tb_frame.f_locals
#             stacks.append(dict(filename=filename,
#                                etype=f'{etype.__name__}', emsg=emsg,
#                                name=name, lineno=lineno, line=line))
#             stack_vars.append(local_vars)
#             if code is not None and line in code:
#                 elineno = lineno
#                 eline = line.strip()
#         tb = tb.tb_next
#     results = dict(etype=f'{etype.__name__}',
#                    emsg=emsg, traceback=list(stacks[::-1]))
#     if code is not None:
#         results['code'] = code
#     if include_locals:
#         results['locals'] = list(stack_vars[::-1])
#     if include_translated:
#         results.update(_defaultErrorModel.get_slots(emsg))
#     if logging_json is not None:
#         logging_json(
#             type='exception_hook',
#             code=code, emsg=emsg,
#             traceback=results['traceback'],
#             elineno=elineno, eline=eline
#         )
#     return results



import builtins
from IPython.display import display, HTML
from IPython import get_ipython
from .timeout import exec_with_timeout
# from kogi.exception_hook import SHOW_TRACEBACK, SHOW_SYNTAXERROR
from kogi.dialog import kogi_catch

_lines = None
_outputs = None


def input_for_judge(s=''):
    global _lines
    if _lines is not None:
        if len(_lines) > 0:
            return _lines.pop(0)
        return ''
    return builtins.input(s)


def print_for_judge(*a, **kw):
    if _outputs is not None:
        sep = kw.get('sep', ' ')
        end = kw.get('end', '\n')
        s = sep.join([str(s) for s in a]) + end
        _outputs.append(s)
    else:
        builtins.print(*a, **kw)


def judge(code, data):
    global _lines, _outputs
    problem_id = data['problem_id']
    # global_vars = get_ipython().user_global_ns.copy()
    global_vars = {
        'print': print_for_judge, 'input': input_for_judge,
    }
    try:
        ac = 0
        for i, testcase in enumerate(data['testcases']):
            title = testcase.get('title', f'Case {i+1}')
            inputData = testcase['input']
            outputData = testcase['output']
            _lines = [s for s in inputData.split('\n') if len(s) > 0]
            _outputs = []
            # local_vars = {
            #     'print': print_for_judge, 'input': input_for_judge,
            # }
            exec_with_timeout(code, global_vars, None, 10)
            resultData = ''.join(_outputs)
            ac += 1 if outputData == resultData else 0
            render_result(title, inputData, resultData, outputData)
        render_footer(data)
    #     log(type='atcoder', problem=problem_id, ac=ac, code=code)
    except SyntaxError:
        SHOW_SYNTAXERROR(get_ipython())
        slots = dict(
            problem_id=problem_id,
        )
        kogi_catch(code=code, context=slots)
    except:
        SHOW_TRACEBACK(get_ipython())
        slots = dict(
            problem_id=problem_id,
            vars=global_vars,
        )
        kogi_catch(code=code, context=slots)
    finally:
        _lines = None
        _outputs = None


JUDGE_CSS = '''
<style>
.parent {
  background-color: #edebeb;
  width: 100%;
  //height: 150px;
}
textarea {
  width: 100%; 
  box-sizing: border-box;  /* ※これがないと横にはみ出る */
  //height:120px; 
  font-size: large;
  outline: none;           /* ※ブラウザが標準で付加する線を消したいとき */
  resize: none;
}
.box18{  // ひだり
  //padding: 0.2em 0.5em;
  //margin: 2em 0;
  color: #565656;
  background: #ffeaea;
  background-size: 150%;
  background-repeat: no-repeat;
  background-position: top right;
  background-color:rgba(255,255,255,0.8);
  //box-shadow: 0px 0px 0px 10px #ffeaea;
  border: dashed 2px #ffc3c3;
  //border-radius: 8px;
}
.box16{
    background: repeating-linear-gradient(-45deg, #D5FFB0, #D5FFB0 3px,#ffffff 3px, #ffffff 7px);
}
.box17{
    background-image: repeating-linear-gradient(45deg, rgba(255, 0, 0, .3), rgba(255, 0, 0, .3) 10px, #FFEFF7 10px, #FFEFF7 20px);
}
.box24 {
    position: relative;
    padding: 0.5em 0.7em;
    margin: 2em 0; background: #6f4b3e;
    color: white; font-weight: bold;
}
.box23:after {
    position: absolute;
    content: '';
    top: 100%; left: 30px;
    border: 15px solid transparent;
    border-top: 15px solid #6f4b3e;
    width: 0; height: 0;
}
</style>
'''


def render_header():
    display(HTML(JUDGE_CSS))


JUDGE_HTML = '''
<div class="parent" style="clear:both;">
<h4>{title}</h4>
<pre>{input}</pre>
<div style="float: left; width: 48%; text-align: right;">
<label class="box24" for="input">実行結果</label>
<textarea class="box18" style="height:{height}" readonly>{output}</textarea>
</div>
<div style="float: left; width: 48%; text-align: right;">
<label class="box24" for="outout">正解例</label>
<textarea class="box18" style="height:{height}" readonly>{sample}</textarea>
</div>
</div>
<p style="clear:both;"></p>
'''


def render_result(title, input, result, output):
    n_lines = max(output.count('\n'), result.count('\n'))+1
    data = dict(
        title=title,
        input=input,
        output=result,
        sample=output,
        box='box16' if result == output else 'box17',
        height='240px' if n_lines > 10 else f'{n_lines*24}px'
    )
    display(HTML(JUDGE_CSS))
    display(HTML(JUDGE_HTML.format(**data)))


AC_HTML = '''
<div style="clear:both;">
AtCoderでACを取るためには、<b>制約条件</b>を満たす全ての入力にパスするようにプログラムする必要があります。<br/>
もう一度、確認してから<a href="{url}" target="atcoder">提出</a>しましょう。
</div>
'''


def render_footer(data):
    display(HTML(AC_HTML.format(url=data['url'])))



import traceback
from .content import ICON, JS, CSS
from IPython.display import display, HTML
from kogi.settings import translate_ja, kogi_get

try:
    from google.colab import output
except ModuleNotFoundError:
    output = None

APPEND_JS = '''
<script>
var target = document.getElementById('{target}');
var content = `{html}`;
if(target !== undefined) {{
    target.insertAdjacentHTML('beforeend', content);
    target.scrollTop = target.scrollHeight;
}}
</script>
'''


def append_content(target, html):
    html = html.replace('\\', '\\\\')
    html = html.replace('`', '\\`')
    display(HTML(APPEND_JS.format(target=target, html=html)))


DIALOG_BOT_HTML = '''
<div class="sb-box">
    <div class="icon-img icon-img-left">
        <img src="{icon}" width="60px">
    </div>
    <div class="icon-name icon-name-left">{name}</div>
    <div class="sb-side sb-side-left">
        <div class="sb-txt sb-txt-left">{text}</div>
    </div>
</div>
'''

DIALOG_USER_HTML = '''
<div class="sb-box">
    <div class="icon-img icon-img-right">
        <img src="{icon}" width="60px">
    </div>
    <div class="icon-name icon-name-right">{name}</div>
    <div class="sb-side sb-side-right">
        <div class="sb-txt sb-txt-right">{text}</div>
    </div>
</div>
'''

dialog_count = 0
dialog_target = None


def cc(text):
    if len(text)==0:
        return text
    n_ascii = sum(1 for c in text if ord(c) < 128)
    #print(text, n_ascii, len(text), n_ascii / len(text))
    if (n_ascii / len(text)) < 0.4:  # 日本語
        t = translate_ja(text)
        # print(t)
        if t is not None:
            return f'{text}<br><i>{t}</i>'
    return text


def kogi_display(text, **kwargs):
    if isinstance(text, list):
        text = '<br>'.join(cc(line) for line in text)
    else:
        text = cc(text)

    data = dict(
        text=text,
        icon='kogi-fs8.png',
        name='コギー',
    )
    data.update(kwargs)
    data['icon'] = ICON(data.get('icon', '/'))
    _HTML = kwargs.get('html', DIALOG_BOT_HTML)
    html = _HTML.format(**data)
    #print('@@', html, kwargs)
    if dialog_target is not None:
        append_content(dialog_target, html=html)
    else:
        display(HTML(CSS('dialog.css') + html))


DIALOG_HTML = '''
<div id="dialog">
    {script}
    <div id="{target}" class="box" style="height: 150px">
    </div>
    <div style="text-align: right">
        <textarea id="input" placeholder="{placeholder}"></textarea>
    </div>
</div>
'''


class Conversation(object):
    slots: dict
    records: list

    def __init__(self, slots=None):
        self.slots = {} if slots is None else slots
        self.records = []

    def get(self, key, value):
        return self.slots.get(key, value)

    def ask(self, input_text):
        output_text = self.response(input_text)
        self.records.append((input_text, output_text))
        return output_text

    def response(self, input_text):
        return 'わん'


def display_dialog(context=None, placeholder='質問はこちらに'):
    global dialog_target, dialog_count
    dialog_target = f'output{dialog_count}'
    dialog_count += 1
    data = dict(
        script=JS('dialog.js'),
        placeholder=placeholder,
        target=dialog_target,
    )
    DHTML = DIALOG_HTML.replace('150', str(kogi_get('chat_height', 180)))
    display(HTML(CSS('dialog.css') + DHTML.format(**data)))
    if context is None:
        context = Conversation()

    def dialog_bot(bot_text, **kwargs):
        nonlocal context
        data = dict(
            icon=context.get('bot_icon', 'kogi-fs8.png'),
            name=context.get('bot_name', 'コギー'),
        )
        data.update(kwargs)
        kogi_display(bot_text, **data)

    def dialog_user(user_text, **kwargs):
        nonlocal context
        data = dict(
            icon=context.get('user_icon', 'girl_think-fs8.png'),
            name=context.get('user_name', 'あなた'),
            html=DIALOG_USER_HTML,
        )
        data.update(kwargs)
        kogi_display(user_text, **data)

    if output is not None:
        def ask(user_text):
            try:
                user_text = user_text.strip()
                dialog_user(user_text)
                bot_text = context.ask(user_text)
                dialog_bot(bot_text)
                #print('@', bot_text)
            except:
                kogi_display('バグで処理に失敗しました。ごめんなさい')
                traceback.print_exc()
        output.register_callback('notebook.ask', ask)

    return dialog_bot, dialog_user




from IPython.core.magic import register_cell_magic
from .atcoder_impl import print, input, _run_judge


@register_cell_magic
def problem(option, code):
    _run_judge(code, option)


@register_cell_magic
def atcoder(option, code):
    _run_judge(code, option)



from IPython.display import display, HTML
#import difflib
import requests
from bs4 import BeautifulSoup
import builtins
from .logger import log
from .liberr import KogiError


_lines = None
_outputs = None


def input(s=''):
    global _lines
    if _lines is not None and len(_lines) > 0:
        return _lines.pop(0)
    else:
        _lines = None
    return builtins.input(s)


def print(*a, **kw):
    if _outputs is not None:
        sep = kw.get('sep', ' ')
        end = kw.get('end', '\n')
        s = sep.join([str(s) for s in a]) + end
        _outputs.append(s)
    else:
        builtins.print(*a, **kw)


def _get_problemid(problem):
    problem, _, _ = problem.partition('&')
    _, _, problem = problem.rpartition('/')
    problem = problem.lower()
    num = 'a'
    if problem[-1] in 'abcdefgh':
        num = problem[-1]
        problem = problem[:-1].replace('_', '').replace('-', '')
    return f'{problem}_{num}'


def _get_url(problem):
    if '/' in problem:
        return problem
    problem = problem.lower()
    num = 'a'
    if problem[-1] in 'abcdefgh':
        num = problem[-1]
        problem = problem[:-1].replace('_', '').replace('-', '')
    return f'https://atcoder.jp/contests/{problem}/tasks/{problem}_{num}'


SAMPLE = {}


def _get_sample(problem):
    pid = None
    try:
        if '/' in problem:
            problem = problem[problem.rfind('/')+1:]
        problem = problem.lower()
        if '_' in problem:
            problem, num = problem.split('_')
        else:
            num = problem[-1].lower()
            problem = problem[:-1].lower()
        pid = f'{problem}_{num}'
        if pid in SAMPLE:
            return SAMPLE[pid], pid
        response_text = requests.get(
            url=f"https://atcoder.jp/contests/{problem}/tasks/{problem}_{num}").text
        html = BeautifulSoup(response_text, "lxml")
        d = {}
        for a in html.find_all("section"):
            # print(a)
            if a.h3 and a.pre:
                key = a.h3.text.replace('\r\n', '\n')
                value = a.pre.text.replace('\r\n', '\n')
                d[key] = value
        SAMPLE[pid] = d
        return d, pid
    except:
        return {}, pid


# def _check_atcoder(option):
#     try:
#         if 'atcoder' in option:
#             d = _get_sample(option)
#             return True
#     except Exception as e:
#         print('問題が読み込めません', e)
#     return False


# _COLOR_HTML_DIC = {
#     'yellow': '<span style="background: pink">',
#     'red': '<span style="background: lightblue">',
#     'end': '</span>'
# }


# def _display_diff(ground_truth, target):
#     _CDIC = _COLOR_HTML_DIC

#     if ground_truth == target:
#         return

#     d = difflib.Differ()
#     diffs = d.compare(ground_truth, target)

#     result = ''
#     for diff in diffs:
#         status, _, character = list(diff)
#         if status == '-':
#             character = _CDIC['red'] + character + _CDIC['end']
#         elif status == '+':
#             character = _CDIC['yellow'] + character + _CDIC['end']
#         else:
#             pass
#         result += character

#     display(
#         HTML(f'<h4>差分</h4><div style="white-space: pre-wrap;">{result}</div>'))


JUDGE_CSS = '''
<style>
.parent {
  background-color: #edebeb;
  width: 100%;
  //height: 150px;
}
textarea {
  width: 100%; 
  box-sizing: border-box;  /* ※これがないと横にはみ出る */
  //height:120px; 
  font-size: large;
  outline: none;           /* ※ブラウザが標準で付加する線を消したいとき */
  resize: none;
}
.box18{  // ひだり
  //padding: 0.2em 0.5em;
  //margin: 2em 0;
  color: #565656;
  background: #ffeaea;
  background-size: 150%;
  background-repeat: no-repeat;
  background-position: top right;
  background-color:rgba(255,255,255,0.8);
  //box-shadow: 0px 0px 0px 10px #ffeaea;
  border: dashed 2px #ffc3c3;
  //border-radius: 8px;
}
.box16{
    background: repeating-linear-gradient(-45deg, #D5FFB0, #D5FFB0 3px,#ffffff 3px, #ffffff 7px);
}
.box17{
    background-image: repeating-linear-gradient(45deg, rgba(255, 0, 0, .3), rgba(255, 0, 0, .3) 10px, #FFEFF7 10px, #FFEFF7 20px);
}
.box24 {
    position: relative;
    padding: 0.5em 0.7em;
    margin: 2em 0; background: #6f4b3e;
    color: white; font-weight: bold;
}
.box23:after {
    position: absolute;
    content: '';
    top: 100%; left: 30px;
    border: 15px solid transparent;
    border-top: 15px solid #6f4b3e;
    width: 0; height: 0;
}
</style>
'''

JUDGE_HTML = '''
<div class="parent" style="clear:both;">
<h4>{title}</h4>
<pre>{input}</pre>
<div style="float: left; width: 48%; text-align: right;">
<label class="box24" for="input">実行結果</label>
<textarea class="box18" style="height:{height}" readonly>{output}</textarea>
</div>
<div style="float: left; width: 48%">
<label class="box24" for="outout">正解例</label>
<textarea class="box18" style="height:{height}" readonly>{sample}</textarea>
</div>
</div>
<p style="clear:both;"></p>
'''

AC_HTML = '''
<div style="clear:both;">
AtCoderでACを取るためには、<b>制約条件</b>を満たす全ての入力にパスするようにプログラムする必要があります。<br/>
もう一度、確認してから<a href="{url}" target="atcoder">提出</a>しましょう。
</div>
'''


def _run_judge(code, problem):
    global _lines, _outputs
    d, problem_id = _get_sample(problem)
    if len(d) == 0:
        raise KogiError(
            translated='問題データが読み込めません。',
            reason='問題の指定方法が間違っています',
            hint='問題ページのURLをコピーしてください',
            solution='%%atcoder 問題ページのURL'
        )
    get_ipython()._run_cell_context = problem_id
    try:
        ac = 0
        display(HTML(JUDGE_CSS))
        for key in ['入力例 1', '入力例 2', '入力例 3']:
            if key not in d:
                continue
            data = {'title': key, 'input': d[key]}
            _lines = [s for s in d[key].split('\n') if len(s) > 0]
            _outputs = []
            res = get_ipython().run_cell(code)
            res.raise_error()
            key = key.replace('入力', '出力')
            data['sample'] = d[key]
            data['output'] = ''.join(_outputs)
            data['box'] = 'box16' if data['sample'] == data['output'] else 'box17'
            ac += 1 if data['sample'] == data['output'] else 0
            lines = max(data['output'].count('\n'),
                        data['sample'].count('\n'))+1
            data['height'] = '240px' if lines > 10 else f'{lines*24}px'
            display(HTML(JUDGE_HTML.format(**data)))
        display(HTML(AC_HTML.format(url=_get_url(problem))))
        get_ipython()._run_cell_context = None
        log(type='atcoder', problem=problem_id, ac=ac, code=code)
    except:
        pass
    finally:
        _lines = None
        _outputs = None



import traceback
from kogi.logger import kogi_print
from .drill import kogi_judge, judge_cpc
from .atcoder import download_atcoder_problem


def atcoder_detector(directive, raw_cell):
    if 'https://atcoder.jp/contests/' in directive:
        return 'atcoder'
    return None


def atcoder_judge(ipy, raw_cell, directive):
    data = download_atcoder_problem(directive)
    if 'error' in data:
        kogi_print(data['error'])
    elif 'problem_id' in data:
        kogi_print('コギーがAtCoderの問題を発見し、テストケースを実行しようとしています')
        kogi_judge(ipy, raw_cell, data, judge_cpc)
    else:
        kogi_print('問題が見つかりません。')
    return None



import time
import traceback
from .content import ICON, JS, CSS
from IPython.display import display, HTML, JSON
from kogi.settings import kogi_log, translate, model_generate, print_nop

try:
    from google.colab import output as colab_output
except ModuleNotFoundError:
    colab_output = None

RMT_HTML = '''
<div id="{id}" class="parent">
<div style="float: left; width: 48%; text-align: right;">
<label class="box24" for="input">{input}</label>
<textarea id="input" class="box16"></textarea>
</div>
<div style="float: left; width: 48%; text-align: right;">
<label class="box24" for="outout">{output}</label>
<textarea id="output" class="box18" readonly></textarea>
</div>
</div>
'''


def display_rmt(input='入力', output='出力', delay=600):
    data = dict(
        id=1,
        input=input,
        output=output,
        delay=600,
    )
    DHTML = RMT_HTML.format(**data)
    display(HTML(CSS('_rmt.css')+DHTML+JS('_rmt.js')))
    return id


class TransformWeaver(object):
    def before(self, text):
        return text

    def after(self, text):
        return text


TransformWeaverNone = TransformWeaver()

MODULES = [
    ('math.', 'import math'),
    ('os.', 'import os'),
    ('sys.', 'import sys'),
    ('collections.', 'import collections'),
    ('time.', 'import time'),
    ('datetime.', 'import datetime'),
    ('random.', 'import random'),
    ('np.', 'import numpy as np'),
    ('scipy.', 'import scipy'),
    ('plt.', 'import matplotlib.pyplot as plt'),
    ('pd.', 'import pandas as pd'),
]


def check_module(code, fmt='<tt>{}</tt> してから、'):
    for module, todo in MODULES:
        if code.startswith(module):
            return fmt.format(todo)
    return ''


def codegen(code, text=''):
    if '<nl>' in code:
        code = code.replace('<nl>', '\n').replace('<tab>', '\t')
    if text != '' or not text.isascii():
        text = translate(text, lang='ja_en')
        text = f'# {text}\n'
    code = check_module(code, fmt='{}\n') + text + code
    return code


_LOGS = []


def rmt(input='入力', output='予測', delay=600, print=print_nop,
        transform=TransformWeaverNone, generate=model_generate):
    display_rmt(input=input, output=output, delay=delay)

    cached = {'': ''}

    def convert(text):
        global _LOGS
        try:
            ss = []
            for line in text.splitlines():
                if line not in cached:
                    if line.isascii():
                        _line = translate(line, lang='en_ja')
                        _line = transform.before(_line)
                        print(line, '=>', _line)
                    else:
                        _line = transform.before(line)
                    s = time.time()
                    _translated = generate(_line)
                    e = time.time()
                    _translated = codegen(_translated, _line)
                    translated = transform.after(_translated)
                    print(f'{e-s:.3f}', line, '=>', translated)
                    if line == _line and translated == _translated:
                        _LOGS.append((line, translated))
                    else:
                        _LOGS.append((line, translated, _line, _translated))
                    cached[line] = translated
                else:
                    translated = cached[line]
                ss.append(translated)
            text = '\n'.join(ss)
            return JSON({'result': text})
        except:
            traceback.print_exc()

    def logger():
        global _LOGS
        if len(_LOGS) > 0:
            kogi_log('rmt', pairs=_LOGS)
            _LOGS = []

    if colab_output is not None:
        colab_output.register_callback('notebook.Convert', convert)
        colab_output.register_callback('notebook.Logger', logger)



import signal


def handler(signum, frame):
    #print("Timeout is over!")
    raise TimeoutError("Time Limit Exceeded")


signal.signal(signal.SIGALRM, handler)


class Timeout(object):
    def __init__(self, timeout=5):
        self.timeout = timeout

    def __enter__(self):
        signal.alarm(self.timeout)

    def __exit__(self, exc_type, exc_value, traceback):
        signal.alarm(0)  # cancel it


def eval_with_timeout(code, globals=None, locals=None, timeout=10):
    with Timeout(timeout):
        return eval(code, globals, locals)


def exec_with_timeout(code, globals=None, locals=None, timeout=10):
    with Timeout(timeout):
        exec(code, globals, locals)



import os
import base64


_cache = {}


def load(file):
    file = f'{os.path.dirname(os.path.abspath(__file__))}/{file}'
    try:
        if file.endswith('.png'):
            file_data = open(file, "rb").read()
            return 'data:image/png;base64,' + base64.b64encode(file_data).decode('utf-8')
        else:
            with open(file) as f:
                return f.read()
    except:
        print(file)
        return


def ICON(file):
    if '/' in file:
        return file
    if file not in _cache:
        _cache[file] = load(file)
    return _cache[file]


def CSS(file):
    assert file.endswith('.css')
    if file not in _cache:
        _cache[file] = load(file)
    data = _cache[file]
    return f'<style>{data}</style>'


def JS(file):
    assert file.endswith('.js')
    if file not in _cache:
        _cache[file] = load(file)
    data = _cache[file]
    return f'<script>{data}</script>'



import os

def download_from_google_drive(file_id, filename='model.zip', model_path = './model', quiet=True):
    os.system('pip install --upgrade gdown')
    import gdown
    url = f'https://drive.google.com/uc?id={file_id}'

    gdown.download(url, file_id, quiet=quiet)
    if file_id.endswith('.zip'):
        os.system(f'rm -rf {model_path}')
        os.system(f'unzip -d {model_path} -j {filename}')




import traceback
from functools import wraps

from kogi.logger import sync_lazy_loggger

from .dialog import kogi_catch
from IPython.core.interactiveshell import InteractiveShell, ExecutionResult


RUN_CELL = InteractiveShell.run_cell
SHOW_TRACEBACK = InteractiveShell.showtraceback
SHOW_SYNTAXERROR = InteractiveShell.showsyntaxerror


DETECTOR = []
RUNNER = {}


def kogi_register_hook(key, runner, detector):
    if key is not None and runner is not None:
        RUNNER[key] = runner
    if detector is not None:
        DETECTOR.append(detector)


def kogi_run_cell(ipy, raw_cell, kwargs):
    directive = None
    result = None
    if '# kogi' in raw_cell:
        _, _, cmd = raw_cell.partition('# kogi')
        directive, _, _ = cmd.partition('\n')
    if directive is None and '#kogi' in raw_cell:
        _, _, cmd = raw_cell.partition('#kogi')
        directive, _, _ = cmd.partition('\n')
    if directive is not None:
        directive = directive.strip()
        for detector in DETECTOR:
            key = detector(directive, raw_cell)
            if key in RUNNER:
                result = RUNNER[key](ipy, raw_cell, directive)
                if not isinstance(result, ExecutionResult):
                    result = RUN_CELL(ipy, 'pass', **kwargs)
                return result
    if result is None:
        result = RUN_CELL(ipy, raw_cell, kwargs)
    return result


def change_run_cell(func):
    @wraps(func)
    def run_cell(*args, **kwargs):
        try:
            #args[1] is raw_cell
            return kogi_run_cell(args[0], args[1], kwargs)
        except:
            traceback.print_exc()
        value = func(*args, **kwargs)
        return value
    return run_cell


def change_showtraceback(func):
    @wraps(func)
    def showtraceback(*args, **kwargs):
        # print('** new version ***')
        # value = func(*args, **kwargs)
        try:
            ipyshell = args[0]
            code = ipyshell.user_global_ns['In'][-1]
            kogi_catch(code=code)
        except:
            traceback.print_exc()
    return showtraceback


def enable_kogi_hook():
    InteractiveShell.run_cell = change_run_cell(RUN_CELL)
    InteractiveShell.showtraceback = change_showtraceback(SHOW_TRACEBACK)
    InteractiveShell.showsyntaxerror = change_showtraceback(SHOW_SYNTAXERROR)


def disable_kogi_hook():
    InteractiveShell.run_cell = RUN_CELL
    InteractiveShell.showtraceback = SHOW_TRACEBACK
    InteractiveShell.showsyntaxerror = SHOW_SYNTAXERROR




import torch
import torch.nn as nn
from torch import Tensor
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import DataLoader

from torch.nn import (TransformerEncoder, TransformerDecoder,
                      TransformerEncoderLayer, TransformerDecoderLayer)
import math
import logging
import hashlib

from kogi.libnmt.downloader import download_from_google_drive

DEVICE = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
UNK_IDX, PAD_IDX, SOS_IDX, EOS_IDX = 0, 1, 2, 3


def get_transform(tokenizer, transform_text=None, max_length=128, max_target_length=None):
    if max_target_length is None:
        max_target_length = max_length

    def encode(src, max_length=max_length):
        inputs = tokenizer.encode(src, max_length=max_length,
                                  add_special_tokens=False, truncation=True, return_tensors='pt')
        input_ids = inputs[0] + 4
        return torch.cat((torch.tensor([SOS_IDX]),
                          input_ids,
                          torch.tensor([EOS_IDX]))).to(dtype=torch.long)

    def decode(output_ids):
        output_ids = [idx-4 for idx in output_ids if idx > 4]
        return tokenizer.decode(output_ids)

    if transform_text is None:
        def transform(src, tgt):
            inputs = encode(src, max_length=max_length)
            targets = encode(tgt, max_length=max_target_length)
            return inputs, targets
    else:
        def transform(src, tgt):
            src, tgt = transform_text(src, tgt)
            inputs = encode(src, max_length=max_length)
            targets = encode(tgt, max_length=max_target_length)
            return inputs, targets
    return encode, decode, transform


# from morichan

class Seq2SeqTransformer(nn.Module):
    def __init__(self,
                 num_encoder_layers: int,
                 num_decoder_layers: int,
                 emb_size: int,
                 nhead: int,
                 src_vocab_size: int,
                 tgt_vocab_size: int,
                 dim_feedforward: int = 512,
                 dropout: float = 0.1):
        super(Seq2SeqTransformer, self).__init__()
        encoder_layer = TransformerEncoderLayer(d_model=emb_size, nhead=nhead,
                                                dim_feedforward=dim_feedforward)
        self.transformer_encoder = TransformerEncoder(
            encoder_layer, num_layers=num_encoder_layers)
        decoder_layer = TransformerDecoderLayer(d_model=emb_size, nhead=nhead,
                                                dim_feedforward=dim_feedforward)
        self.transformer_decoder = TransformerDecoder(
            decoder_layer, num_layers=num_decoder_layers)

        self.generator = nn.Linear(emb_size, tgt_vocab_size)
        self.src_tok_emb = TokenEmbedding(src_vocab_size, emb_size)
        self.tgt_tok_emb = TokenEmbedding(tgt_vocab_size, emb_size)
        self.positional_encoding = PositionalEncoding(
            emb_size, dropout=dropout)

    def forward(self,
                src: Tensor,
                tgt: Tensor,
                src_mask: Tensor,
                tgt_mask: Tensor,
                src_padding_mask: Tensor,
                tgt_padding_mask: Tensor,
                memory_key_padding_mask: Tensor):
        src_emb = self.positional_encoding(self.src_tok_emb(src))
        tgt_emb = self.positional_encoding(self.tgt_tok_emb(tgt))
        memory = self.transformer_encoder(src_emb, src_mask, src_padding_mask)
        outs = self.transformer_decoder(tgt_emb, memory, tgt_mask, None,
                                        tgt_padding_mask, memory_key_padding_mask)
        return self.generator(outs)

    def encode(self, src: Tensor, src_mask: Tensor):
        return self.transformer_encoder(self.positional_encoding(
            self.src_tok_emb(src)), src_mask)

    def decode(self, tgt: Tensor, memory: Tensor, tgt_mask: Tensor):
        return self.transformer_decoder(self.positional_encoding(
            self.tgt_tok_emb(tgt)), memory,
            tgt_mask)


class PositionalEncoding(nn.Module):
    def __init__(self,
                 emb_size: int,
                 dropout: float,
                 maxlen: int = 5000):
        super(PositionalEncoding, self).__init__()
        den = torch.exp(- torch.arange(0, emb_size, 2)
                        * math.log(10000) / emb_size)
        pos = torch.arange(0, maxlen).reshape(maxlen, 1)
        pos_embedding = torch.zeros((maxlen, emb_size))
        pos_embedding[:, 0::2] = torch.sin(pos * den)
        pos_embedding[:, 1::2] = torch.cos(pos * den)
        pos_embedding = pos_embedding.unsqueeze(-2)

        self.dropout = nn.Dropout(dropout)
        self.register_buffer('pos_embedding', pos_embedding)

    def forward(self, token_embedding: Tensor):
        return self.dropout(token_embedding +
                            self.pos_embedding[:token_embedding.size(0), :])


class TokenEmbedding(nn.Module):
    def __init__(self, vocab_size: int, emb_size):
        super(TokenEmbedding, self).__init__()
        self.embedding = nn.Embedding(vocab_size, emb_size)
        self.emb_size = emb_size

    def forward(self, tokens: Tensor):
        return self.embedding(tokens.long()) * math.sqrt(self.emb_size)

# モデルが予測を行う際に、未来の単語を見ないようにするためのマスク


def generate_square_subsequent_mask(sz):
    mask = (torch.triu(torch.ones((sz, sz), device=DEVICE)) == 1).transpose(0, 1)
    mask = mask.float().masked_fill(mask == 0, float(
        '-inf')).masked_fill(mask == 1, float(0.0))
    return mask

# ソースとターゲットのパディングトークンを隠すためのマスク
# モデルが予測を行う際に、未来の単語を見ないようにするためのマスク


def create_mask(src, tgt):
    src_seq_len = src.shape[0]
    tgt_seq_len = tgt.shape[0]

    tgt_mask = generate_square_subsequent_mask(tgt_seq_len)
    src_mask = torch.zeros((src_seq_len, src_seq_len),
                           device=DEVICE).type(torch.bool)

    src_padding_mask = (src == PAD_IDX).transpose(0, 1)
    tgt_padding_mask = (tgt == PAD_IDX).transpose(0, 1)
    return src_mask, tgt_mask, src_padding_mask, tgt_padding_mask

# train/eval


def collate_fn(batch):
    src_batch, tgt_batch = [], []
    for src_ids, tgt_ids in batch:
        src_batch.append(src_ids)
        tgt_batch.append(tgt_ids)
    src_batch = pad_sequence(src_batch, padding_value=PAD_IDX)
    tgt_batch = pad_sequence(tgt_batch, padding_value=PAD_IDX)
    return src_batch, tgt_batch


def train(train_iter, model, batch_size, loss_fn, optimizer):
    model.train()
    losses = 0

    # 学習データ
    #collate_fn = string_collate(hparams)
    train_dataloader = DataLoader(
        train_iter, batch_size=batch_size, shuffle=True,
        collate_fn=collate_fn, num_workers=2)

    for src, tgt in train_dataloader:
        src = src.to(DEVICE)
        tgt = tgt.to(DEVICE)

        tgt_input = tgt[:-1, :]

        src_mask, tgt_mask, src_padding_mask, tgt_padding_mask = create_mask(
            src, tgt_input)

        logits = model(src, tgt_input, src_mask, tgt_mask,
                       src_padding_mask, tgt_padding_mask, src_padding_mask)

        optimizer.zero_grad()

        tgt_out = tgt[1:, :]
        loss = loss_fn(
            logits.reshape(-1, logits.shape[-1]), tgt_out.reshape(-1))
        loss.backward()

        optimizer.step()
        losses += loss.item()

    return losses / len(train_dataloader)


def evaluate(val_iter, model, batch_size, loss_fn):
    model.eval()
    losses = 0

    val_dataloader = DataLoader(
        val_iter, batch_size=batch_size, shuffle=True,
        collate_fn=collate_fn, num_workers=2)

    for src, tgt in val_dataloader:
        src = src.to(DEVICE)
        tgt = tgt.to(DEVICE)

        tgt_input = tgt[:-1, :]

        src_mask, tgt_mask, src_padding_mask, tgt_padding_mask = create_mask(
            src, tgt_input)

        logits = model(src, tgt_input, src_mask, tgt_mask,
                       src_padding_mask, tgt_padding_mask, src_padding_mask)

        tgt_out = tgt[1:, :]
        loss = loss_fn(
            logits.reshape(-1, logits.shape[-1]), tgt_out.reshape(-1))
        losses += loss.item()

    return losses / len(val_dataloader)


# greedy search を使って翻訳結果 (シーケンス) を生成
# https://kikaben.com/transformers-evaluation-details/#chapter-2

def _greedy_decode(model, src, src_mask, max_len, start_symbol, device):  # original
    src = src.to(device)
    src_mask = src_mask.to(device)

    memory = model.encode(src, src_mask)
    ys = torch.ones(1, 1).fill_(start_symbol).type(torch.long).to(device)
    for i in range(max_len-1):
        memory = memory.to(device)
        tgt_mask = (generate_square_subsequent_mask(ys.size(0))
                    .type(torch.bool)).to(device)
        out = model.decode(ys, memory, tgt_mask)
        out = out.transpose(0, 1)
        prob = model.generator(out[:, -1])
        _, next_word = torch.max(prob, dim=1)
        next_word = next_word.item()

        ys = torch.cat([ys,
                        torch.ones(1, 1).type_as(src.data).fill_(next_word)], dim=0)
        if next_word == EOS_IDX:
            break
    return ys

# 翻訳

def md5(filename):
    with open(filename, 'rb') as f:
        return hashlib.md5(f.read()).hexdigest()

def save_model(hparams, model, file='transformer-model.pt'):
    torch.save(dict(
        tokenizer=hparams.tokenizer_name_or_path,
        additional_tokens=hparams.additional_tokens,
        num_encoder_layers=hparams.num_encoder_layers,
        num_decoder_layers=hparams.num_decoder_layers,
        emb_size=hparams.emb_size,
        nhead=hparams.nhead,
        vocab_size=hparams.vocab_size + 4,
        fnn_hid_dim=hparams.fnn_hid_dim,
        model=model.state_dict(),
    ), file)
    logging.info(f'saving... {file} {md5(file)}')


def load_model(filename, AutoTokenizer, device='cpu', dynamic_qint8=False):
    if isinstance(device, str):
        device = torch.device(device)
    checkpoint = torch.load(filename, map_location=device)
    tokenizer = AutoTokenizer.from_pretrained(checkpoint['tokenizer'])
    tokenizer.add_tokens(checkpoint['additional_tokens'])
    model = Seq2SeqTransformer(
        checkpoint['num_encoder_layers'],
        checkpoint['num_decoder_layers'],
        checkpoint['emb_size'],
        checkpoint['nhead'],
        checkpoint['vocab_size'],
        checkpoint['vocab_size'],
        checkpoint['fnn_hid_dim']
    )
    model.load_state_dict(checkpoint['model'])

    if dynamic_qint8:
        model = torch.quantization.quantize_dynamic(
            model, {torch.nn.Linear}, dtype=torch.qint8
        )
    model.to(device)
    return model, tokenizer

def load_transformer_nmt(filename, device='cpu', dynamic_qint8=False, print=print):
    if not filename.endswith('.pt'):
        download_from_google_drive(filename, filename='model.pt')
        filename='model.pt'
    if isinstance(device, str):
        device = torch.device(device)
    
    model, tokenizer = load_model(filename, device, dynamic_qint8)
    encode, decode, _ = get_transform(tokenizer)

    def generate_greedy(src: str, beam=1, max_length=128) -> str:
        model.eval()
        src = encode(src).view(-1, 1).to(device)
        num_tokens = src.shape[0]
        src_mask = (torch.zeros(num_tokens, num_tokens)).type(torch.bool)
        output_ids = _greedy_decode(
            model, src, src_mask, max_len=max_length, start_symbol=SOS_IDX, device=device).flatten()
        return decode(output_ids)
    return generate_greedy



from IPython.display import display, HTML
from IPython.core.magic import register_line_magic

# https://docs.google.com/presentation/d/e/2PACX-1vR96clROL-GKBYq5LwREjOmpryFHLULrbaikjEs3-CecbuDqMFO9-MAZPD2_VY6SWiJZMsRaHPeWjCV

URL = 'https://docs.google.com/presentation/d/e/{}'

SLIDE_HTML = '''
<iframe src="{}/embed?start=false&loop=true&delayms=3000" 
    frameborder="0" width="720" height="430" 
    allowfullscreen="true" mozallowfullscreen="true" webkitallowfullscreen="true">
</iframe>
'''


@register_line_magic
def slide(line):
    if not line.startswith('https://'):
        line = URL.format(line)
    display(HTML(SLIDE_HTML.format(line)))
